#!/usr/local/perl -w
#
# DISAMBIGUATION PROGRAM USING CONCEPTUAL DENSITY
# TREE MODULE
#
# Author: Eneko Agirre Bengoa
#         Ixa Taldea NLP group
#         Computer Languages and Systems Department
#         Basque Country University
#
# Copyright (c) Eneko Agirre 1998
#
# No redistribution without prior permission. Published work containing
# results derived from use of this software must contain an appropriate
# acknowledgement.
#
# interface:
#    
# Parameters (defined in WsdDens.pm) :
#   WsdDens::ALPHA
#   WsdDens::BETA
#   WsdDens::PARTOF
#   WsdDens::LOCALMU
#   WsdDens::GLOBALMU
#   WsdDens::WORDS

package Tree ;
use WNaccess qw(kb_concept2pos kb_word2concepts kb_all_ascendants kb_all_hypernyms
	        kb_descendants kb_local_mu) ;
use Dens qw(dens_node) ;

use Exporter () ;
@ISA = qw(Exporter) ;
@EXPORT = qw(tree_init tree_init_counter 
	     tree_compute_dist tree_compute_dist_hyp
	     tree_compute_dens tree_select_max_concept tree_select_next_mac_concept
	     tree_mark_disambiguated tree_explain_disambiguated tree_return_disambiguated
	     tree_select_next_max_concept tree_subsumedwords tree_delete_cycles
	     WORDS SENSES_BEFORE SENSES_AFTER STILL_AMBIGUOUS MONOSEMOUS NOT_IN_WN) ;

use strict ;

########
#
# data structure holding nodes in KB under consideration 
# with all desired information
#
# encoding of concepts, if WN used: synset#pos
#
# tree = {
#   tops  => {}          : set of concepts which are tops (implemented as a hash)
#   node  => {}          : hash with info for each node in subtree
#   words => { set of {  : set of words to be disambiguated (implemented as a hash)
#              $word => { set of      : word
#                         $concept }  : set of concepts (implemented as a hash)
#                                       e.g. $TREE{words}{$word}{$mainconcept} = 1
#      }
#   amb_words => {}      : set of ambiguous words (implemented as hash)
#   senses => $          : number of senses of words to be disambiguated
#   }
#
# WHERE node is a structure containing:
#
# {
#  sw => {}           : set of words under concept (implemented as hash)
# #  c =>              : concept code
#  h => {}           : set of hyponym concepts (implemented as hash)
#  y => {}           : set of hypernym concepts (implemented as hash)
#  d =>              : depth of concept (minimum if several parents)
#  w =>              : density value
#                      AFTER DISAMBIGUATION CAN BE DE WEIGHT OF THE SELECTED CONCEPT
# #  sr =>             : synonymy ratio (? ONLY in compute-synonymy-intern)
# #  o =>              : sum of overlaping ratio: 1/[number of times synset added per word]*overlaping
# #  s =>              : number of hyponyms (? ONLY in compute-s-intern)
# #  r =>              : sum of relevances: either length of sw OR sum of the relevances
#                      provided in insert-tree-sense-intern.
#  e =>              : true if selected concept (or concetp is ancestor/descendant of a selected concept)
#  x =>              : list of disambiguated concepts which are below (value only if e = true)
#  b =>              : true if bottom synset
# # m =>               : true if concept is an input word (or is ancestor of input word)
#  

use vars '%TREE' ; # global variable


my($WORDS) = 0;
my($SENSES_BEFORE) = 0 ;  
my($SENSES_AFTER) = 0 ;
my($STILL_AMBIGUOUS) = 0 ;
my($MONOSEMOUS) = 0 ;
my($NOT_IN_WN) = 0 ;


sub tree_init_counter () {
  $WORDS = 0;
  $SENSES_BEFORE = 0 ;  
  $SENSES_AFTER = 0 ;
  $STILL_AMBIGUOUS = 0 ;
  $MONOSEMOUS = 0 ;
  $NOT_IN_WN = 0 ;
}

sub tree_init ($@) {
  my($pos,@list) = @_ ;
  my($entry,$word,$concept,$w) ;
  %TREE = (
	   tops  => {},
	   node  => {},
	   words => {},
	   senses => 0 
	  ) ;
  foreach $entry (@list) {
    if ($entry =~ /^(.+?)#(.+)$/) {
	($word,$concept) = ($1,$2) ;
	tree_add_word($word,kb_concept2pos($concept),$concept) or print STDERR "Tree::tree_init: tree_add_word failed $word#$concept\n"; }
    else {
      $word = $entry ; 
      tree_add_word($word,$pos) or print STDERR "Tree::tree_init: tree_add_word failed $word\n"; }
  }
}

# 
# tree_add_word
# input: $word: word to include in tree
#        $pos:  pos of the word
#     OPTIONAL
#        $mainconcept: specific concept of word         
#        $relevance
# parameter:
#        $WsdDens::PARTOF if defined and true, use all ascendants,
#        else use only hypernymy.
# output: 1 if successful, 0 otherwise
# process: if word already in %TREE do nothing
#          update %TREE, inserting either all concepts of the word (with given pos)
#          OR the single concept provided. this involves:
#          - insert those concepts
#          - insert all ascendants of those
#          - include the new word in $TREE->{words}{$word}
#          - update number of senses of words
#        
# LISP: wn-tree.lisp::compute-tree-sense-intern
#
sub tree_add_word ($$;$$) {
  my($word,$pos,$mainconcept,$relevance) = @_;
  
  my(@concepts, $concept) ;
  my($pathset,$path) ;

  if (defined($mainconcept)) {
    return if defined($TREE{words}{$word}{$mainconcept}) ;
    @concepts = ($mainconcept) ;}
  else {
    return if defined($TREE{words}{$word}) ;
    @concepts = kb_word2concepts($word,$pos) ; }

  if (!scalar(@concepts)) { 
    print STDERR "Tree::tree_add_word: unknown word $word ($pos)\n"; 
    undef $TREE{words}{$word} ;
    return 
  };

  # update TREE: new word with its concepts
  $TREE{senses} += scalar(@concepts) ;  

  foreach $mainconcept (@concepts) {
    $TREE{words}{$word}{$mainconcept} = 1 ;
    if ($WsdDens::PARTOF) {
      $pathset = kb_all_ascendants($mainconcept) ; }
    else {
      $pathset = kb_all_hypernyms($mainconcept) ; }

    die "Tree::tree_add_word: incorrect concept\n" if !scalar(@{ $pathset }) ;

    foreach $path (@{ $pathset }) {
      # update TREE: tops
      $TREE{tops}{ ${ $path }[0] } = 1 ;
      
      tree_insert_concept($word,$mainconcept,$path) ;
    }
  }
  # put ambiguous words in $TREE{amb_words}
  if (defined($TREE{amb_words}{$word})) {
    delete $TREE{amb_words}{$word} if scalar(keys %{ $TREE{words}{$word} }) == 1 ; }
  else {
    $TREE{amb_words}{$word} = 1 if scalar(keys %{ $TREE{words}{$word} }) > 1 ; }
  return 1 ;
}

#
# tree_insert_concept
#
# input: $word:        inserted word
#        $mainconcept: concept of inserted word
#        $path:        hypernymy (or ancestors) path (passed as reference to array)
#    OPTIONAL
#        $hypernym:    hypernym of first concept in path (undefined by default)
#        $depth:       depth of concept (1 by default)
# 
# output: 1
# effect: updated TREE (node)
#
# process: create a node for each concept in the path, with the fol. info:
#          - sw: add $word to the set of words
#          - d: put given depth (d is the minimum of all depths)
#          - b: 1 if bottom leaf (has a word below), 0 otherwise.
#          - y: list of hypernyms
#          - h: list of hyponyms
#
# LISP: wn-tree.lisp::insert-tree-sense

sub tree_insert_concept ($$$;$$) {
  my($word,$mainconcept,$path,$hypernym,$depth) = @_ ;
  my($concept);
  my($ref) ;

  $depth = 1 if !defined($depth) ;
  print "Tree::tree_insert_concept: ($word,$mainconcept,$$path[0])\n" if $main::DEBUG;

  return 1 if !scalar(@{ $path }) ;
  $concept = shift @{ $path } ;

  if (defined($TREE{node}{$concept})) {
    $ref = $TREE{node}{$concept} }
  else {
    $ref = $TREE{node}{$concept} = {} ;
  }

  $ref->{sw}{$word} = 1 ;

  if (!defined($ref->{d})) {
    $ref->{d} = $depth ; }
  else {
    $ref->{d} = $depth if ($ref->{d} > $depth) ; }
  
  $ref->{'y'}{$hypernym} = 1 if defined($hypernym) ;
  if (!scalar(@{ $path })) {
    $ref->{b} = 1; }
  else {
    $ref->{b} = 0; 
    $ref->{h}{${ $path }[0]} = 1 ;
    tree_insert_concept($word,$mainconcept,$path,$concept,$depth+1) ;
  }
  (1) ;
}

#
# computing density on the tree
#
#
#
# external parameters:
# $WsdDens::ALPHA        # which density function
# $WsdDens::ALPHA        # alpha in density function
# $WsdDens::BETA         # beta in density function
# $WsdDens::LOCALMU      # use global mu or local mu (mu = average number of hyponyms)
# $WsdDens::GLOBALMU     # in WN 1.4 
# $WsdDens::WORDS

sub tree_compute_dens () {
  if (! defined($WsdDens::FUNC)) {
    $WsdDens::FUNC = ".dens" ;
  }
  foreach ($WsdDens::FUNC) {
    /dens/           && return(tree_compute_dens_orig()) ;
    /int/            && return(tree_compute_dens_orig()) ;
    /areas/          && return(tree_compute_dens_orig()) ;
    /height/         && return(tree_compute_dens_orig()) ;
    /links/          && return(tree_compute_dens_rec()) ;
    /dist/           && return(tree_compute_dens_rec()) ;
    die "unknown density function: $WsdDens::FUNC\n" ;
  }
}

my %VIS ;
sub tree_compute_dens_orig () {
  my($top,@hypos,$hypo) ;
  undef %VIS ;
  foreach $top (keys %{ $TREE{tops} }) {
    @hypos = ($top) ;
    while ($hypo = shift(@hypos)) {
      if ($VIS{$hypo}) {
	next ;
      }
      $VIS{$hypo} = 1 ;
      tree_node_dens($hypo, $TREE{node}{$hypo}) ;
      push @hypos, keys %{ $TREE{node}{$hypo}->{h} };
    }
  }
}

sub tree_node_dens ($$) {
  my($concept,$ref) = @_ ;
  my($descendants,$marks,$sum,$mu,$descmarks,$h) = (0,0,0,0,0,0) ;
  my($m) ;

  $descendants = kb_descendants($concept) ;
  $marks = scalar(keys %{ $ref->{sw} }) ;
  if ($WsdDens::LOCALMU) {
    $mu = kb_local_mu($concept) ; }
  else {
    $mu = $WsdDens::GLOBALMU ;}

  foreach ($WsdDens::FUNC) {
    /areas/ && do {
      foreach $m (keys %{ $ref->{sw} }) {
	$descmarks += kb_descendants($m) ; } } ;
    /height/ && do {
      $h = kb_height($concept) ; }
  }
  
  $ref->{w} = dens_node($descendants,$mu,$marks,$WsdDens::ALPHA,$WsdDens::BETA) ;

  print "> $concept: $marks $descendants $mu $ref->{w}\n" if $main::DEBUG;
}


#
# computing distance on the tree only if one is hypernym or synonym of other
# looks for concepts with both words below 
# computes their weight, and returns the node with less distance from those 
# that have one concept over the other
#
# output:
#   - the selected concept
#   - the distance

sub tree_compute_dist_hyp ($$) {
  my($ca,$cb) = @_ ;
  my($node,$winner,$dist,$top) ;

  $dist = 99999 ;
  foreach $top (keys %{ $TREE{tops} }) {
    next if keys(%{$TREE{node}{$top}{sw}}) < 2 ;
    tree_distance_node($top) ;
  }

  foreach $node (keys %{ $TREE{node} }) {
    next if ! grep /$node/, ($ca,$cb) ;
    next if keys(%{$TREE{node}{$node}{sw}}) < 2 ;
    next if $TREE{node}{$node}{w} > $dist ;
    $winner = $node ;
    $dist = $TREE{node}{$node}{w} ;
  }
  if ($dist < 99999) {
    return($dist) }
  else {
    return }
      
}

#
# computing distance on the tree
# looks for concepts with both words below 
# computes their weight, and returns the node with less distance
# 
# output:
#   - the selected concept
#   - the distance

sub tree_compute_dist ($$) {
  my($ca,$cb) = @_ ;
  my($node,$winner,$dist,$top) ;

  $dist = 99999 ;
  foreach $top (keys %{ $TREE{tops} }) {
    next if keys(%{$TREE{node}{$top}{sw}}) < 2 ;
    tree_distance_node($top) ;
  }

  foreach $node (keys %{ $TREE{node} }) {
    next if keys(%{$TREE{node}{$node}{sw}}) < 2 ;
    next if $TREE{node}{$node}{w} > $dist ;
    $winner = $node ;
    $dist = $TREE{node}{$node}{w} ;
  }
  if ($dist < 99999) {
    return($dist) }
  else {
    return }
      
}

# 
# compute distance for the node above two concepts
# for each hyponym add 
#   the inverse of the depth of each node (each link)
#   plus all weights below
# if no hyponyms, distance is zero

sub tree_distance_node ($) {
  my($node) = @_ ;
  my($dist, $hypo) ;

  $dist = 0 ;
  foreach $hypo (keys %{ $TREE{node}{$node}->{h} }) {
    $dist += tree_distance_node($hypo)
      + 1/$TREE{node}{$node}{d}
    }
  $TREE{node}{$node}->{w} = $dist ;
}



#
# returns the winning concept
#
sub tree_select_max_concept () {
  my($bestw,$bestc) = (0) ;
  my($concept,$ref,$overlap,$word) ;

  while (($concept,$ref) = each %{ $TREE{node} }) {
    next if $ref->{e} ;
    next if scalar( keys %{ $ref->{sw} } ) <= 1 ;
    next if $ref->{w} < $bestw ;
    # concept disambiguates still ambiguous words
    $overlap = 0;
    foreach $word (keys %{ $ref->{sw} }) {
      if (defined($TREE{amb_words}{$word})) {
	$overlap = 1;
	last ; }
    }
    next if !$overlap ;
    $bestw = $ref->{w} ;
    $bestc = $concept ;
  }
  return($bestc) ;
}

#
# returns the next winning concept
#
sub tree_select_next_max_concept (;$) {
  my($sel_concept) = @_ ;
  my(@diswords) ;

  if (defined($sel_concept)) {
    # mark tree so as not to select same concept (or concepts below)
    @diswords = keys %{ $TREE{node}{$sel_concept}->{sw} } ;
    tree_mark_hypernyms($sel_concept, \@diswords) ;
    tree_mark_hyponyms($sel_concept, \@diswords,$TREE{node}{$sel_concept}->{w} ) ;
  }

  # return max concept from those not marked
  return(tree_select_max_concept()) ;
}


sub tree_mark_disambiguated ($) {
  my($sel_concept) = @_ ;
  my(@diswords,$word,$concept) ;

  @diswords = keys %{ $TREE{node}{$sel_concept}->{sw} } ;
  tree_mark_hypernyms($sel_concept, \@diswords) ;
  tree_mark_hyponyms($sel_concept, \@diswords,$TREE{node}{$sel_concept}->{w} ) ;
  
  # delete disambiguated words from list of ambiguous words
  # delete senses of disambiguated words not selected 
  foreach $word (@diswords) {
    delete $TREE{amb_words}{$word} ;
    foreach $concept (keys %{ $TREE{words}{$word} }) {
      tree_delete_unselected_senses($concept,\@diswords) }}

}

# 
# sets the field that stores which nouns are disambiguated below this one
#
sub tree_mark_hypernyms ($$) {
  my($concept,$diswords_ref) = @_ ;

  print "Tree::tree_mark_hypernyms: ($concept)\n" if $main::DEBUG;  
  foreach (@{ $diswords_ref }) {
    $TREE{node}{$concept}->{'x'}{$_} = 1 ; }
  foreach (keys %{ $TREE{node}{$concept}->{'y'} }) {
    tree_mark_hypernyms ($_, $diswords_ref) ; }
}

#
# sets the field (e) which marks that a concept above has been selected
# sets the weight (w) to that of the selected concept
# sets the field (x) that which nouns were disambiguated by the selected concept above
#
sub tree_mark_hyponyms ($$$) {
  my($concept,$diswords_ref,$winning_weight) = @_ ;
  my($ref) = $TREE{node}{$concept} ;

  print "Tree::tree_mark_hyponyms: ($concept)\n" if $main::DEBUG;  
  $ref->{w} = $winning_weight if !$ref->{e}; # put weight of selected concept (ONCE)
  $ref->{e} = 1 ;
  foreach (@{ $diswords_ref}) {
    $ref->{'x'}{$_} = 1 ; }
  foreach (keys %{ $ref->{h} }) {
    tree_mark_hyponyms ($_, $diswords_ref,$winning_weight) ; }
}

sub tree_delete_unselected_senses($$) {
  my($concept,$diswords_ref) = @_ ;

  my($ref) = $TREE{node}{$concept} ;

  # return if {e} is set or
  #           @diswords and ->{'x'} have any word in common
  return if $ref->{e} ;
  foreach (@{ $diswords_ref }) {
    return if $ref->{'x'}{$_} } ;
  
  print "Tree::tree_delete_unselected_senses: ($concept)\n" if $main::DEBUG;  

   # delete disambiguated words from here
  foreach (@{ $diswords_ref }) {
    delete $ref->{sw}{$_} } ;

  # recompute density for this concept
  if (0 < scalar keys %{ $ref->{sw} }) {
    tree_node_dens($concept,$ref) } 
  else {
    $ref->{w} = 0 ; }

  # treat hypernyms too
  foreach (keys %{ $ref->{'y'} }) {
    tree_delete_unselected_senses($_,$diswords_ref) }
}

sub tree_explain_disambiguated (@) {
  my(@selected_concepts) = @_;
  my($word,$c) ;

  print STDERR "Sel concepts: " . join(" ",@selected_concepts) . "\n" ;
  foreach $c (@selected_concepts) { tree_pprint_node($c) } ;

  print STDERR "Disambiguated senses: \n" ;
  foreach $word (keys %{ $TREE{words} } ) {
    $WORDS++ ;
    foreach $c (keys %{ $TREE{words}{$word} }) {
      $SENSES_BEFORE++ ;
      if ($TREE{node}{$c}->{e}) {
	$SENSES_AFTER++ ;
	tree_pprint_node($c);}}}

  print STDERR "Still ambiguous words: " ;
  foreach $word (keys %{ $TREE{amb_words} }) {
    $STILL_AMBIGUOUS++;
    print STDERR "$word " ; }
  print STDERR "\n" ;


  print STDERR "Monosemous words: " ;
  foreach $word (keys %{ $TREE{words} } ) {
    if (scalar keys %{ $TREE{words}{$word} } == 1) {
      $MONOSEMOUS++;
      print STDERR "$word " ; }}
  print STDERR "\n" ;

  print STDERR "Words not in WordNet: " ;
  foreach $word (keys %{ $TREE{words} } ) {
    if ((scalar keys %{ $TREE{words}{$word} }) == 0) {
      $MONOSEMOUS++;
      print STDERR "$word " ; }}
  print STDERR "\n" ;

  
}

sub tree_pprint_node ($) {
  my($concept) = @_ ;
  my($e,$b) ;
  
  my($ref) = $TREE{node}{$concept} ;
  $e = (defined($ref->{e}) && $ref->{e}) ? "Y" : "N" ;
  $b = (defined($ref->{b}) && $ref->{b}) ? "Y" : "N" ;
  print STDERR "-- $concept: weight($ref->{w}) depth($ref->{d}) desc(" . kb_descendants($concept) . ") selected($e) bottom($b)\n" ;
  print STDERR "               sw(words below) " . join(" ",keys %{ $ref->{sw} }) . "\n" ;
  print STDERR "               x(selected words below)  " . join(" ",keys %{ $ref->{'x'} }) . "\n" if $ref->{'x'};
}

sub tree_print () {
  foreach (keys %{ $TREE{node} }) {
    tree_pprint_node($_) ; }}

#
# result is hash of words and concepts, where value is 0 if monosmeous or the weight 
# of the selected concept disambiguating the concept:
#     $result_ref{$word}{$c} = X
#
sub tree_return_disambiguated () {
  my($word,$c) ;
  my($result_ref) ;

  $result_ref = {} ;

  # add disambiguated words
  foreach $word (keys %{ $TREE{words} } ) {
    foreach $c (keys %{ $TREE{words}{$word} }) {
      if ($TREE{node}{$c}->{e}) {
	${$result_ref}{$word}{$c} = $TREE{node}{$c}->{w} ;  }
      else {
	${$result_ref}{$word}{$c} = 0 }
      }}


  # add monosemous
  foreach $word (keys %{ $TREE{words} } ) {
    if (scalar keys %{ $TREE{words}{$word} } == 1) {
      ($c) = keys %{ $TREE{words}{$word} } ;
      ${ $result_ref }{$word}{$c} = 1 }}

#  # return completely ambiguous words too
#  foreach $word (keys %{ $TREE{amb_words} } ) {
#      ${ $result_ref }{$word}{'X'} = "X" }

  return($result_ref) ;
}


# returns the list of nouns subsumed by the given concept
sub tree_subsumedwords ($) {
  my($concept) = @_ ;

  return(keys %{ $TREE{node}{$concept}->{sw} });
}



#
# deletes any cycle in tree
#
sub tree_delete_cycles () {
  my(@hypos,$top,$hypo) ;
  foreach $top (keys %{ $TREE{tops} }) {
    tree_delete_cycles_aux ($top,"") ;
  }
}

sub  tree_delete_cycles_aux ($$) {
  my ($c,$string)=@_ ;
  my ($hypo,$parent) ;
  if ($string =~ /$c/) {
#    print STDERR "cycle in: $string $c\n"; 
    ($parent) = reverse split(/ /,$string) ;
    delete $TREE{node}{$c}{"y"}{$parent} ;
    delete $TREE{node}{$parent}{h}{$c} ;
    return }
  foreach $hypo (keys %{ $TREE{node}{$c}->{h} }) {
    tree_delete_cycles_aux($hypo, "$string $c");
  }
}

#
# output tree in a linear way
#
sub tree_linearize () {
  my(@hypos,$top,$hypo) ;
  foreach $top (keys %{ $TREE{tops} }) {
    tree_linearize_aux ($top,"") ;
  }
  return 0 ;
}

sub tree_linearize_aux ($$) {
  my ($c,$string)=@_ ;
  my ($hypo) ;
  print "$string $c" ;
  if ($string =~ /$c/) {
    print "**\n"; 
    return }
  else {
    print "\n"}
  foreach $hypo (keys %{ $TREE{node}{$c}->{h} }) {
    tree_linearize_aux($hypo, "$string $c");
  }
}
(1) 
