#!/local/perl5.005_03/bin/perl -w 
#
# WORDNET ACCESS SYSTEM FOR PERL
#
# Author: Eneko Agirre Bengoa
#         Ixa Taldea NLP group
#         Computer Languages and Systems Department
#         Basque Country University
#
# Copyright (c) Eneko Agirre 1998
#
# No redistribution without prior permission. Published work containing
# results derived from use of this software must contain an appropriate
# acknowledgement.
#
# MODULE: WNacess.pm
#
#         accessing WN 1.6 index and data lines
#
# Compilation:
#    COMPILE_ONTOLOGY: function to be executed ONCE before using WordNet 
#
# Interface:
#    kb_init_ontology	initilization
#    kb_hyponyms        given a concept returns its hyponyms
#    kb_all_ascendants	yields all ascendnts following hyper/meronymy
#    kb_all_hypernyms   yields all ascendants followin hypernymy
#    kb_word2concepts 	given a word (+ an optional POS) returns its concepts
#    kb_concept2pos	given a concept returns its POS
#    kb_concept2sf	given a concept returns its semantic file number
#    kb_concept2words   given a concept return the words in the synset
#    kb_concept2senses  given a concept return the senses in the synset 
#    kb_sense2concept   given a sense   return the concept
#    kb_descendants     given a concept return the number of descendants
#    kb_local_mu        given a concept return the local average number of hyponyms
#
# Concepts and senses:
#    concept = "pos_letter#synset_number" e.g. "n#02364995"
#    sense   = "pos#word#sense_number"    e.g. "n#car#5"
#
# Implementation issues:
#    WordNet indexes: they are compiled into static hashes for random access
#                     (SEE $wn_sense_index_tied and compile_ontology)
#
#    WordNet data files: they are left untouched
#
#    Other files:        synset2senseno
#                        senseno2file
#                        senseno2synset
#                        desc.n.tie || desc.pof.n.tie || desc.words.n.tie
#
#    PARAMETERS (defined in WsdDens.pm)
#       WsdDens::PARTOF  "" || ".pof" 
#       WsdDens::WORDS   "" || ".words"            
#    they condition with descendant file is used

package WNaccess ;
use Dens ;
use FileHandle ;
use Fcntl ;
use DB_File ;
use Exporter () ;
@ISA = qw(Exporter) ;
@EXPORT = qw(kb_init_ontology kb_all_ascendants kb_all_hypernyms
	     kb_hyponyms
	     kb_concept2pos kb_word2concepts kb_concept2sf kb_concept2words
	     kb_sense2concept kb_concept2senses
	     kb_descendants kb_local_mu) ;

use strict ;
use vars qw(%INDEXF %INDEXFILE %DATAF %DATAFILE %DESC %INDEX %DATA
	    $wn_dict_path $wn_gloss_index $wn_sense_index $wn_lexnames
	    $wn_sense_index_tied $dict_path $wn_synset2senseno $wn_senseno2file
	    $wn_senseno2synset  $wn_descendantsfile $wn_descendantsfiletied) ;

#
# wordnet index and data files
#

if (defined($ENV{WNHOME})) {
  $wn_dict_path  = "$ENV{WNHOME}/dict" ; }
elsif (defined($ENV{WNSEARCHDIR})) {
  $wn_dict_path  = "$ENV{WNSEARCHDIR}" ; }
else {
  die "WNaccess.pm: wordnet paths not installed\n" ; }

if (! -e "$wn_dict_path/index.noun") {
  die "WNaccess.pm: wordnet not properly installed\n" }

print STDERR "> Using WordNet at $wn_dict_path\n" ;
 
%INDEXF = (n => "$wn_dict_path/index.noun" ,
	   a => "$wn_dict_path/index.adj" ,
	   v => "$wn_dict_path/index.verb" ,
	   r => "$wn_dict_path/index.adv" ) ;

%DATAF = (n => "$wn_dict_path/data.noun" ,
	  a => "$wn_dict_path/data.adj" ,
	  v => "$wn_dict_path/data.verb" ,
	  r => "$wn_dict_path/data.adv" ) ;

$wn_gloss_index = "$wn_dict_path/index.gloss" ;
$wn_sense_index = "$wn_dict_path/index.sense" ;

$wn_lexnames = "$wn_dict_path/lexnames" ;

#
# opens filehandles for each datafile into %DATAFILE
#

my($pos) ;
foreach $pos (keys %DATAF) {
  $DATAFILE{$pos} = new FileHandle "< $DATAF{$pos}" || die $!;
}
  

#
# other preprocessed files (see initialization)
#


#
# sense.index coding of sysnet_type (1 n noun, 2 v verb, 3 a adj, 4 r adv, 5 s satellite adj) ;
# 
# there are two parallel codings for pos:
#   synset_type: n v a r s
#   pos:         n v a r
#
my @synset_type = ("","n","v","a","r","s") ;
my @pos_list = ("n","v","a","r") ;


#############################
#
# COMPILATION
#
# creates a file with tied sense.index enabling random access
# creates a tied descendants file for random access
#
sub kb_compile_ontology () {
  kb_init_ontology() ;
}

#
# tie INDEX.SENSE to a hash for quick access
#
sub kb_compile_sense_index () {
  my ($word,$i) ;
  unlink $wn_sense_index_tied if (-e $wn_sense_index_tied) ;
  open(I,$wn_sense_index) || die $! ;
  tie %INDEXFILE, "DB_File", $wn_sense_index_tied , O_RDWR|O_CREAT, 0666, $DB_HASH or die $!;
  print STDERR "$wn_sense_index -> creating tied file $wn_sense_index_tied...\n" ;

  # store each line
  while (<I>) {
    ($word) = (/^(.*?)%/) ;
    if (! defined($INDEXFILE{$word})) {
      $INDEXFILE{$word} = $_  ; }
    else {
      $INDEXFILE{$word} .= $_ ; }
    
    print "$word\n$INDEXFILE{$word}\n" if $main::DEBUG ; 
  }
  close(I) ;
  untie %INDEXFILE ;
  print STDERR "done\n" ;
}

#############################
#
# INITIALIZATION
#
# %INDEXFILE
# %DESC
#
# two possibilities:
#    load all INDEX.SENSE and DATA files in hash structures (too expensive!!)
#   OR
#    use tied version of INDEX.SENSE and use open filehandles for DATA files
#    call to get_sense_index and get_data_line every time before using the data structures
#
# SECOND PREFERED!
#
# yields:
#    INDEX{lemma}{pos}{sensenumber} = {
#         stype                            : like pos, but with satellite also
#         lexfile                          : lexicographic file code (number)
#         lexid                            : sense number in lexfile
#         synset                           : synset number
#
#    DATA{pos}{synsetoffset} = {
#         stype                            : like pos, but with satellite also
#         lexfile                          : lexicographic file code (number)
#         words [ list of {                : words in the synset
#           word                           : words
#           lexid                          : indentifies sense within lexfile
#           } ]
#         pointers                         : hash of pointers to other synsets
#           {ptr_type} [ list of {         : type of pointer, holding a list of 
#             stype                        : like pos, but with satellite also 
#             synset                       : synset offset
#             pos                          : pos of pointer synset
#             sourtar                      : distiguishes lexical and semantic pointers
#           } ]
#
# in WN 1.6 the memory requirements are approximately:
#    INDEX     6 MB 
#    DATA   16.5 MB (probably less, as the glosses are not stored)
#
#############################


sub kb_init_ontology () {

  $dict_path =  (getpwnam("jirhizts"))[7] . "/Mrd/WordNet" ; # internal-data
  if (! -d $dict_path) {
    mkdir $dict_path, 0777 || die $! ; 
    print STDERR "WNaccess:kb_init_ontology created the following data path: $dict_path\n" ;
  }
 
  $wn_sense_index_tied = "$dict_path/index.sense.tied" ;
  $wn_synset2senseno  = "$dict_path/synset.senseno" ;
  $wn_senseno2file    = "$dict_path/senseno.file" ;
  $wn_senseno2synset  = "$dict_path/senseno.synset" ;
  $wn_descendantsfiletied = "$dict_path/desc.n" . 
                            $WsdDens::PARTOF . 
                            $WsdDens::WORDS .
                            ".tied" ;

  # compile if necessary

  (-e $wn_sense_index_tied)    || kb_compile_sense_index() ;
  (-e $wn_descendantsfiletied) || kb_compile_descendants_file() ;


  # tie sense index to %INDEXFILE hash for random access
  tie %INDEXFILE, "DB_File", $wn_sense_index_tied , O_RDONLY, 0660
             or die "Cannot open file '$wn_sense_index_tied': $!\n";
  print STDERR "> $wn_sense_index_tied tied\n" ;

  # tie %DESC to file with descendant information
  tie %DESC, "DB_File", $wn_descendantsfiletied , O_RDONLY, 0660 
             or die "Cannot open file '$wn_descendantsfiletied': $!\n";
  print STDERR "> $wn_descendantsfiletied tied\n" ;

  # loads everything in memory
  # takes too long not used yet !

#  load_sense_index() ;
#  load_data_files() ;

  print STDERR "> WordNet kb initialized\n" ;
}

sub load_sense_index () {
  open(I,$wn_sense_index) || die $! ;

  undef %INDEX ;
  # parse each line
  while (<I>) {
    parse_sense_index_line($_) ;
  }
  close(I) ;
}

sub load_data_files () {
  open(I,$DATAF{n}) || die $! ;

  # take out first 30 lines
  my $i = 0 ;
  while (<I>) {$i++; last if $i==29} ; 

  undef %DATA ;
  # parse each line
  while (<I>) {
    parse_data_line($_) ;
  }
  close(I) ;
}

#
# input:  word
# output: 1 if word is known to WordNet, 0 otherwise
# sideeffect: clear %INDEX, and stores WN information for word into %INDEX
#
sub get_sense_index ($) {
  my($word) = @_ ;
  my($l) ;
#  die "compiled index.sense not loaded (%INDEXFILE)\n" if !defined(%INDEXFILE) ;
  undef %INDEX ;
  $l = $INDEXFILE{$word} ;
  if (!defined($l)) {
    return(0) ; }
  else {
    foreach (split(/\n/,$l)) {
      parse_sense_index_line($_) ; }
    return(1) ;
  }
}
    


sub parse_sense_index_line ($) {
  my $l = shift ;

  $_ = $l ;

  # get all fields into array
  my ($sensekey,$synsetoffset,$sensenumber,$tagcnt) = split ; 
  my ($lemma,$synset_type_number,$lexfile,$lexid) = split(/[%:]/,$sensekey) ;
  my $synset_type = $synset_type[$synset_type_number] ;
  my $pos = $synset_type ;
  $pos = "a" if $pos =~ /s/ ;
  $INDEX{$lemma}{$pos}{$sensenumber} = { stype => $synset_type, 
					 lexfile => $lexfile, 
					 lexid => $lexid, 
					 synset => $synsetoffset } ; 
  
}
			  

#
# input:  concept "pos#synset" 
# sideeffect: clear %DATA, and stores WN information for concept into %DATA
#
sub get_data_line ($) {
  my($concept) = @_ ;
  my($pos,$offset) = split(/#/,$concept) ;
  die "WNaccess::get_data_line: wrong concept '$concept'\n" if !defined($offset);
  undef %DATA ;			   
  $DATAFILE{$pos}->seek($offset,0) ;
  parse_data_line( $DATAFILE{$pos}->getline) ;
}  

# parses the WN data line and stores the information in %DATA
# output: also returns a reference to the information in %DATA for this synset
#
sub parse_data_line ($) {
  my $l = shift ;
  my(@F, $c, $word, $lex_id) ;
  $_ = $l ;

  # get all fields into array
  @F = split ;
  my ($synset_offset,$lexfile,$synset_type,$word_number) = @F[0..3] ;
  my $pos = $synset_type ;
  $pos = "a" if $synset_type eq "s" ;
  $word_number = hex($word_number) ;
  @F = @F[4..$#F] ;

  my $words = [] ;
  for ($c=0 ; $c < $word_number ; $c++) {
    $word = shift @F;
    # treat possible syntactic marker in adjectives
    $word =~ s/\(.+\)$//o if $pos eq "a" ;
    $lex_id = shift @F;
    push @{ $words },
         { word => $word,
	   lexid => hex($lex_id) 
	 } ;
  }

  my $pointer_number = shift @F ;
  
  my $aux = {} ;
  my ($pointer_symbol , $synset_offset_ptr, $pos_ptr, $source_target) ;
  for ($c=0 ; $c < $pointer_number ; $c++) {
    $pointer_symbol = shift @F;
    $synset_offset_ptr = shift @F;
    $pos_ptr = shift @F;
    $source_target = shift @F;

    if (defined( $aux->{$pointer_symbol})) {
      push @{ $aux->{$pointer_symbol} },
           {  stype => $synset_type,
	      synset => $synset_offset_ptr,
	      "pos" => $pos_ptr, 
	      sourtar => $source_target
	   } ;
    }
    else {
      $aux->{$pointer_symbol} = [
           {  stype => $synset_type,
	      synset => $synset_offset_ptr,
	      "pos" => $pos_ptr, 
	      sourtar => $source_target
	   } ] ;

    }
  }
  # frames for verbs: ignored

  # gloss: ignored

  $DATA{$pos}{$synset_offset} = { synset_type => $synset_type,
				  lexfile => $lexfile,
				  words => $words , 
				  pointers => $aux  
				} ;

  return $DATA{$pos}{$synset_offset} ;

}



#########################
#
# basic access interface
#
#########################

sub kb_hyponyms ($) {
  my ($concept) = @_ ;
  my (@res, %DATA_CPY, $target, ) ; 
  my ($pos,$synset) = split(/\#/,$concept) ;
			    
  die "WNaccess::kb_hyponyms: wrong concept\n" if !defined($pos) || !defined($synset) ;

  $DATA_CPY{$pos}{$synset} = get_data_line($concept) ;
  die "WNaccess::kb_hyponyms: %DATA not initialized\n" if !scalar(%DATA) ;

  foreach $_ (@{ $DATA_CPY{$pos}{$synset}{pointers}{"~"} }) {
      push @res, $_->{"pos"} . "#" . $_->{synset} ;
  }
  return @res ;
}

#
# input: concept
# output: a reference to a list of possible paths
#         each path is a reference to a list of concepts
#         ordered from most general (top) to the input concept
#         following hypernymy and meronymy relations
#
# PARTOF relations are only permited ONCE !!
#
# process: uses %DATA_CPY to store a copy of the info for each concept
#
sub kb_all_ascendants ($) {
  my ($concept) = @_ ;
  my ($res) = [] ;
  my (%DATA_CPY, $rel, $target, $newconcept, $res1, $chain) ; 
  my ($pos,$synset) = split(/#/,$concept) ;
			    
  die "WNaccess::kb_all_ascendants: wrong concept\n" if !defined($pos) || !defined($synset) ;

  $DATA_CPY{$pos}{$synset} = get_data_line($concept) ;
  die "WNaccess::kb_all_ascendants: %DATA not initialized\n" if !scalar(%DATA) ;

  foreach $rel ("@","%m","%s","%p") {
    foreach $target (@{ $DATA_CPY{$pos}{$synset}{pointers}{$rel} }) {
      $newconcept = $target->{"pos"} . "#" . $target->{synset} ;
      if ($rel eq "@") {
	$res1 = kb_all_ascendants($newconcept) ;}
      else {
	$res1 = kb_all_hypernyms($newconcept) ; }

      foreach $chain ( @{ $res1 } ) {
	push @{ $chain }, $concept ;
	push @{ $res }, $chain ;
      }
    }
  }
  if (!defined($res) || !scalar(@{ $res })) {
    $res = [ [ $concept ] ]}
  return $res ;
}


#
# input: concept
# output: a reference to a list of possible paths
#         each path is a reference to a list of concepts
#         ordered from most general (top) to the input concept
#         following hypernymy relations
#
sub kb_all_hypernyms ($) {
  my ($concept) = @_ ;
  my ($res) = [] ;
  my ($target, $res1, $newconcept, $chain) ;
  my ($pos,$synset) = split(/#/,$concept) ;
			    

  die "WNaccess::kb_all_hypernyms: wrong concept\n" if !defined($pos) || !defined($synset) ;

  get_data_line($concept) ;
  die "WNaccess::kb_all_hypernyms: %DATA not initialized\n" if !scalar(%DATA) ;

  foreach $target (@{ $DATA{$pos}{$synset}{pointers}{"@"} }) {
    $newconcept = $target->{"pos"} . "#" . $target->{synset} ;
    $res1 = kb_all_hypernyms($newconcept) ;
    foreach $chain ( @{ $res1 } ) {
      push @{ $chain }, $concept ;
      push @{ $res }, $chain ;
    }
  }

  if (!defined($res) || !scalar(@{ $res })) {
    $res = [ [ $concept ] ]}
  return $res ;
}


############################
#
# encoding/decoding concepts
#
############################

#    kb_concept2pos	given a concept returns its POS
#
sub kb_concept2pos (@) {
  # input: list of concpet "n#45362646" OR "n#"
  # output: list of pos per each concept
  return (split(/#/,shift @_))[0] ;
}		    

		
#    kb_word2concepts 	given a word (+ an optional POS) returns its concepts
#    output: list of concepts
#
sub kb_word2concepts ($;$) {
  my ($word,$pos) = @_ ;
  my(@r, @pos, $key, $info) ;

  die "WN::kb_word2concepts: parameter missing\n" if !(defined($word));
#  die "WN::kb_word2concepts: %INDEX not initialized\n" if !scalar(%INDEX) ;

  if (defined($pos)) {
    @pos = ($pos) }
  else {
    @pos = @pos_list}
  
  get_sense_index($word) ;

  foreach $pos (@pos) {
    while (($key,$info) = each %{ $INDEX{$word}{$pos} }) {
      push @r, "$pos#" . $info->{synset}; } }

  print "WNaccess::kb_word2concepts: $pos $word not found in WordNet\n" if !scalar(@r) ;
  return @r ;
}

#    kb_word2concepts 	given a word (+ an optional POS) returns its semantic files
#    output: list of semantic files
#
sub kb_concept2sf ($) {
  my ($concept) = @_ ;
  my($pos,$synset) = split(/#/,$concept) ;
  my(%DATA_CPY) ;

  $DATA_CPY{$pos}{$synset} = get_data_line($concept) ;
  print "WNaccess::kb_concept2sf: concept not found in WordNet\n" if !defined($DATA_CPY{$pos}{$synset}) ;

  return $DATA_CPY{$pos}{$synset}->{lexfile} ;
}

# return the words in the synset
sub kb_concept2words ($) {
  my ($concept) = @_ ;
  my($pos,$synset) = split(/#/,$concept) ;
  my(%DATA_CPY) ;
  my(@words) = () ;

  $DATA_CPY{$pos}{$synset} = get_data_line($concept) ;
  print "WNaccess::kb_concept2words: concept not found in WordNet\n" if !defined($DATA_CPY{$pos}{$synset}) ;

  foreach (@{ $DATA_CPY{$pos}{$synset}->{words} }) {
    push @words, $_->{word} ;
  }			   
  return @words ;
}

# given a sense (pos#word#sense) return the concept belonging to the sense: 
sub kb_sense2concept ($) {
  my ($sense) = @_ ;
  my($pos,$word,$senseno) = split(/#/,$sense) ;

  get_sense_index($word) || die "kb_sense2concept: word $word not in index\n" ;
  my $r = $INDEX{$word}{$pos}{$senseno}{synset} ;
  defined($r) || die "WNaccess::WNaccess::kb_sense2concept: sense $sense not in index\n"  ;
  return "$pos#$r" ;
}

# given a concept and returns a list with the senses (pos#word#sense)
#
# THIS IS EXPENSIVE IN WORDNET!!
#
sub kb_concept2senses ($) {
  my ($concept) = @_ ;
  my($word,$sense, $sf,$sf_sense) ;
  my(%DATA_CPY) ;
  my(@words) = () ;
  my($pos,$synset) = split(/#/,$concept) ;

  $DATA_CPY{$pos}{$synset} = get_data_line($concept) ;
  print "WNaccess::kb_concept2senses: concept not found in WordNet\n" if !defined($DATA_CPY{$pos}{$synset}) ;

  $sf = $DATA_CPY{$pos}{$synset}->{lexfile} ;
  foreach (@{ $DATA_CPY{$pos}{$synset}->{words} }) {
    $word = $_->{word} ;
    $sf_sense = $_->{lexid} ;
    get_sense_index($word) || die "kb_concepts2sense: word $word not in index\n" ;
    
    foreach $sense (keys %{ $INDEX{$word}{$pos} }) {
      if (($INDEX{$word}{$pos}{$sense}{lexfile} == $sf) and
	  ($INDEX{$word}{$pos}{$sense}{lexid} == $sf_sense)) {
	push @words, "$pos#$word#$sense" ;
	last ;
      }
    }
  }			   
  return @words ;
}



################################
#
# descendants file

#
# store number of descendants and average number of hyponyms (mu) in
# tied hash:
#    $DESC{"n#synset"} = "$desc $mu"
# 

# store which concepts were already visited (cycles)  
my(%VISITED) ;
# store auxiliar information
my(%DESC_AUX) ;
my(%SYNSBELOW) ;

sub kb_compile_descendants_file () {
  my ($synset,$desc,$localmu,$localmu2) ;
  my (@tops, $i, $concept, $k, $v) ;
  my ($mu, $heigth, $heigth_hypo, $syn) ;

  my($cat) = "n" ;
  unlink $wn_descendantsfiletied if (-e $wn_descendantsfiletied) ;

  open(I,$DATAF{$cat}) || die $! ;

  tie %DESC, "DB_File", $wn_descendantsfiletied, O_RDWR|O_CREAT, 0660 or die $!;
  print STDERR "$DATAF{$cat} -> creating tied file $wn_descendantsfiletied...\n" ;

  #
  # 1) get set of tops

  $i = 0 ; 
  while (<I>) {$i++; last if $i==29} ; 
  while (<I>) {next if / @ /; /^(.*?) /; push @tops, "$cat#$1" }
  print STDERR "> $cat TOPS:\n" ;
  print STDERR "> " . join("\n> ",@tops) . "\n" ;
  close(I) ;

  #
  # 2) process each node

  open(I,$DATAF{$cat}) || die $! ;
  foreach $concept (@tops) {
    kb_compute_descendants($concept) ;
    print STDERR "top $concept: $DESC{$concept} $DESC_AUX{$concept}\n" ; }
  close(I) ;
  
  #
  # 3) compute overal mu value
  print "\n" ;

  $heigth = 0 ;
  foreach $concept (@tops) {
    $heigth_hypo = $DESC_AUX{$concept} ;
    foreach $syn (keys %{ $SYNSBELOW{$concept} }) {
      $SYNSBELOW{'n#top'}{$syn} = 1 ;
    }
    $heigth = ($heigth_hypo+1) if ($heigth_hypo+1) > $heigth ; }
  close(I) ;

  $desc = keys %{  $SYNSBELOW{'n#top'} } ;
  print STDERR "TOTAL: $desc $heigth " .
               dens_compute_mu($desc, $heigth) . "\n" ;

  undef %VISITED ;
  undef %DESC_AUX ;
  untie %DESC ;
  print STDERR "done\n" ;
}  


sub kb_compute_descendants ($;$) {
  my($concept,$depth) = @_ ;
  my(@refs, $ref,@hypos,$hyporef,$hypo_concept,$desc_hypo,$heigth_hypo,$mu,$syn,$word) ;
  my($desc,$heigth ) = (1,0,0) ;
  my (%DATA_CPY) ; 
  my ($pos,$synset) = split(/\#/,$concept) ;
			    
  $depth = 1 if !defined($depth) ;
  $VISITED{$concept} = 1 ;

  if ($WsdDens::WORDS) {
    foreach $word (kb_concept2words($concept)) {
      $SYNSBELOW{$concept}{$word} = 1 ; 
    }
  }
  else {
    $SYNSBELOW{$concept}{$concept} = 1 ;
  }

  die "WNaccess::kb_compute_descendants: wrong concept\n" if !defined($pos) || !defined($synset) ;

  print "> $pos#$synset\n" if $main::DEBUG ;
  $DATA_CPY{$pos}{$synset} = get_data_line($concept) ;
  die "WNaccess::kb_compute_descendants: %DATA not initialized\n" if !scalar(%DATA) ;
			    
  @refs = () ;
  push @refs, $DATA_CPY{$pos}{$synset}{pointers}{"~"} ;
#  if ($WsdDens::PARTOF) {
#    push @refs, $DATA_CPY{$pos}{$synset}{pointers}{"#m"}, 
#                $DATA_CPY{$pos}{$synset}{pointers}{"#s"},
#                $DATA_CPY{$pos}{$synset}{pointers}{"#p"} ;
#  }
  @hypos = () ;
  foreach $ref ( @refs ) {
    push @hypos, @{ $ref } if defined($ref)}
  if ($#hypos > -1) {
    foreach $hyporef (@hypos) {
      $hypo_concept = "$pos#$hyporef->{synset}" ;

      # this hyponym was already visited
      #
      if (defined($DESC{$hypo_concept})) {
	$heigth_hypo = $DESC_AUX{$hypo_concept} ;
	foreach $syn (keys %{ $SYNSBELOW{$hypo_concept} }) {
	  $SYNSBELOW{$concept}{$syn} = 1 ;
	}
	$heigth = ($heigth_hypo+1) if ($heigth_hypo+1) > $heigth ; 
	print STDERR "repeated: $concept $hypo_concept\n" if $main::DEBUG ; }

      # this hyponym is in a cycle!
      #
      elsif (defined($VISITED{$hypo_concept})) {
	print STDERR "cycle: $concept $hypo_concept\n" ;  }

      # normal unvisited hypo
      #
      else {
	kb_compute_descendants($hypo_concept,$depth+1) ;
	$heigth_hypo = $DESC_AUX{$hypo_concept} ;
	foreach $syn (keys %{ $SYNSBELOW{$hypo_concept} }) {
	  $SYNSBELOW{$concept}{$syn} = 1 ;
	}
	$heigth = ($heigth_hypo+1) if ($heigth_hypo+1) > $heigth ;
      }
    } 
    $desc = keys %{ $SYNSBELOW{$concept} };
    $mu = dens_compute_mu($desc, $heigth ) ;
  }
  else {  
    $heigth = 1 ;
    $mu = 1 ; }

  $DESC{$concept}="$desc $mu" ;
  $DESC_AUX{$concept}= $heigth ;
}

sub kb_descendants ($) {
#  die "kb_descendants: %DESC hash not initialized\n" if (!scalar(%DESC));	
  $DESC{shift @_} =~ /^(.*) / || die "kb_descendants: unknown concept\n" ;
  return $1 ;
}


sub kb_local_mu ($) {
#  die "kb_local_mu: %DESC hash not initialized\n" if (!scalar(%DESC));
  $DESC{shift @_} =~ / (.*)$/ || die "kb_local_mu: unknown concept\n" ;
  return $1 ;
}



#########################
#
# initialization

kb_init_ontology() ;

(1)



